"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.handlePatch = handlePatch;
exports.isPatch = isPatch;
var _ramda = require("ramda");
function _typeof(o) { "@babel/helpers - typeof"; return _typeof = "function" == typeof Symbol && "symbol" == typeof Symbol.iterator ? function (o) { return typeof o; } : function (o) { return o && "function" == typeof Symbol && o.constructor === Symbol && o !== Symbol.prototype ? "symbol" : typeof o; }, _typeof(o); }
function ownKeys(e, r) { var t = Object.keys(e); if (Object.getOwnPropertySymbols) { var o = Object.getOwnPropertySymbols(e); r && (o = o.filter(function (r) { return Object.getOwnPropertyDescriptor(e, r).enumerable; })), t.push.apply(t, o); } return t; }
function _objectSpread(e) { for (var r = 1; r < arguments.length; r++) { var t = null != arguments[r] ? arguments[r] : {}; r % 2 ? ownKeys(Object(t), !0).forEach(function (r) { _defineProperty(e, r, t[r]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(e, Object.getOwnPropertyDescriptors(t)) : ownKeys(Object(t)).forEach(function (r) { Object.defineProperty(e, r, Object.getOwnPropertyDescriptor(t, r)); }); } return e; }
function _defineProperty(e, r, t) { return (r = _toPropertyKey(r)) in e ? Object.defineProperty(e, r, { value: t, enumerable: !0, configurable: !0, writable: !0 }) : e[r] = t, e; }
function _toPropertyKey(t) { var i = _toPrimitive(t, "string"); return "symbol" == _typeof(i) ? i : i + ""; }
function _toPrimitive(t, r) { if ("object" != _typeof(t) || !t) return t; var e = t[Symbol.toPrimitive]; if (void 0 !== e) { var i = e.call(t, r || "default"); if ("object" != _typeof(i)) return i; throw new TypeError("@@toPrimitive must return a primitive value."); } return ("string" === r ? String : Number)(t); }
function isPatch(obj) {
  return (0, _ramda.has)('__dash_patch_update', obj);
}
function getLocationIndex(value, previous) {
  if ((0, _ramda.is)(Number, value) && value < 0) {
    return previous.length + value;
  }
  return value;
}
function getLocationPath(location, obj) {
  var current = [];
  for (var i = 0; i < location.length; i++) {
    var value = getLocationIndex(location[i], (0, _ramda.path)(current, obj));
    current.push(value);
  }
  return current;
}
var patchHandlers = {
  Assign: function Assign(previous, patchOperation) {
    var params = patchOperation.params,
      location = patchOperation.location;
    return (0, _ramda.assocPath)(location, params.value, previous);
  },
  Merge: function Merge(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, _objectSpread(_objectSpread({}, prev), patchOperation.params.value), previous);
  },
  Extend: function Extend(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.concat)(prev, patchOperation.params.value), previous);
  },
  Delete: function Delete(previous, patchOperation) {
    return (0, _ramda.dissocPath)(patchOperation.location, previous);
  },
  Insert: function Insert(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.insert)(getLocationIndex(patchOperation.params.index, prev), patchOperation.params.value, prev), previous);
  },
  Append: function Append(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.append)(patchOperation.params.value, prev), previous);
  },
  Prepend: function Prepend(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.prepend)(patchOperation.params.value, prev), previous);
  },
  Add: function Add(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, prev + patchOperation.params.value, previous);
  },
  Sub: function Sub(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, prev - patchOperation.params.value, previous);
  },
  Mul: function Mul(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, prev * patchOperation.params.value, previous);
  },
  Div: function Div(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, prev / patchOperation.params.value, previous);
  },
  Clear: function Clear(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.empty)(prev), previous);
  },
  Reverse: function Reverse(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, (0, _ramda.reverse)(prev), previous);
  },
  Remove: function Remove(previous, patchOperation) {
    var prev = (0, _ramda.path)(patchOperation.location, previous);
    return (0, _ramda.assocPath)(patchOperation.location, prev.filter(function (item) {
      return !(0, _ramda.equals)(item, patchOperation.params.value);
    }), previous);
  }
};
function handlePatch(previousValue, patchValue) {
  var reducedValue = previousValue;
  for (var i = 0; i < patchValue.operations.length; i++) {
    var patch = patchValue.operations[i];
    patch.location = getLocationPath(patch.location, reducedValue);
    var handler = patchHandlers[patch.operation];
    if (!handler) {
      throw new Error("Invalid Operation ".concat(patch.operation));
    }
    reducedValue = handler(reducedValue, patch);
  }
  return reducedValue;
}