"use strict";

var _index = require("../actions/index");
var _paths = require("../actions/paths");
var _stores = require("./stores");
/**
 * Set the props of a dash component by id or path.
 *
 * @param idOrPath Path or id of the dash component to update.
 * @param props The props to update.
 */
function set_props(idOrPath, props) {
  var ds = (0, _stores.getStores)();
  for (var y = 0; y < ds.length; y++) {
    var _ds$y = ds[y],
      dispatch = _ds$y.dispatch,
      getState = _ds$y.getState;
    var componentPath = void 0;
    var _getState = getState(),
      paths = _getState.paths;
    if (!Array.isArray(idOrPath)) {
      componentPath = (0, _paths.getPath)(paths, idOrPath);
    } else {
      componentPath = idOrPath;
    }
    dispatch((0, _index.updateProps)({
      props: props,
      itempath: componentPath,
      renderType: 'clientsideApi'
    }));
    dispatch((0, _index.notifyObservers)({
      id: idOrPath,
      props: props
    }));
  }
}

// Clean url code adapted from https://github.com/braintree/sanitize-url/blob/main/src/constants.ts
// to allow for data protocol.
var invalidProtocols = /^([^\w]*)(javascript|vbscript)/im;
var newLines = /&(tab|newline);/gi;

// eslint-disable-next-line no-control-regex
var ctrlChars = /[\u0000-\u001F\u007F-\u009F\u2000-\u200D\uFEFF]/gim;
var htmlEntities = /&#(\w+)(^\w|;)?/g;
var clean_url = function clean_url(url) {
  var fallback = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'about:blank';
  if (url === '') {
    return url;
  }
  var cleaned = url.replace(newLines, '').replace(ctrlChars, '').replace(htmlEntities, function (_, dec) {
    return String.fromCharCode(dec);
  }).trim();
  if (invalidProtocols.test(cleaned)) {
    return fallback;
  }
  return url;
};
var dc = window.dash_clientside = window.dash_clientside || {};
dc['set_props'] = set_props;
dc['clean_url'] = dc['clean_url'] === undefined ? clean_url : dc['clean_url'];